package j8086emu.controller;

import j8086emu.model.hardware.CPU;
import j8086emu.model.hardware.registers.RegNames;
import j8086emu.model.progcode.*;
import j8086emu.model.progcode.files.SaveLoadManager;
import j8086emu.model.utils.Properties;
import j8086emu.view.GUI;
import j8086emu.view.TextWidget;

import javax.swing.*;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author vlad
 */
public class ButtonsController {

    private ProgramCode programCode;
    private RunProgramHandler runProgramHandler;
    private SaveLoadManager saveLoadManager;
    private JTextArea codeArea;
//    private JTextArea executionArea;
    private JTable execTable;
    private RamController ramController;
    private DosBoxConfigWriter dosBoxConfigWriter;
    private TasmListingParser tasmListingParser;
    private String codeFilePath = Properties.EXTERNAL_PROGS_PATH + Properties.CODE_FILENAME;
    private GUI gui;

    private void setConfig() {
        dosBoxConfigWriter.write();
    }

    public ButtonsController(JTextArea cArea, JTable execTable, HashMap widMap, RamController ramController, GUI gui) {
        this.gui = gui;
        this.ramController = ramController;
        this.codeArea = cArea;
        //this.executionArea = execArea;
        this.execTable = execTable;
        this.programCode = new ProgramCode();
        PointerCode pointerCode = new PointerCode();
        CPU cpu = ramController.getCpu();

        setWidgetsToHardware(cpu, widMap);
        this.runProgramHandler = new RunProgramHandler(programCode, pointerCode, cpu, gui);
        saveLoadManager = new SaveLoadManager(programCode, pointerCode);
        cpu.getInstructionsMap().setRunProgramHandler(this.runProgramHandler);

        dosBoxConfigWriter = new DosBoxConfigWriter();
        setConfig();
        tasmListingParser = new TasmListingParser();
    }

    public void compile() {
        //saveLoadManager.saveInside(codeArea.getText());
        saveLoadManager.saveOutside(codeFilePath, codeArea.getText());
        tasmListingParser.resetListing();
        runProgramHandler.compile();
        tasmListingParser.parseListing();
        if(!checkAndShowErrorMessages()) {
            runProgramHandler.disassembleProgram(tasmListingParser.getMachineInstructions());
            gui.showMachineInstructions(tasmListingParser.getMachineInstructions());
        }
    }

    // Show error and warning messages and if exist errors block execution buttons and return true
    private boolean checkAndShowErrorMessages() {
        ArrayList<String> errorMessages = tasmListingParser.getErrorMessages();
        ArrayList<String> warningMessages = tasmListingParser.getWarningMessages();

        gui.showErrorWarningMessages(errorMessages, warningMessages);

        if (!errorMessages.isEmpty()) {
            gui.setEnabledExecutionButtons(false);
            return true;
        } else {
            gui.setEnabledExecutionButtons(true);
            return false;
        }
    }

    public void stepByStep() {
        saveLoadManager.saveInside(codeArea.getText());
        refreshExecutionCode();
        try {
            runProgramHandler.runStep();
        } catch (Exception ex) {
            Logger.getLogger(ButtonsController.class.getName()).log(Level.SEVERE, null, ex);
        }
        ramController.visualizeStack();
    }

    public void runAllProgram() {
        saveLoadManager.saveInside(codeArea.getText());
        refreshExecutionCode();
        try {
            runProgramHandler.runAllProgram();
        } catch (Exception ex) {
            Logger.getLogger(ButtonsController.class.getName()).log(Level.SEVERE, null, ex);
        }
        ramController.visualizeStack();
    }

    public void save(String selectedFile) {
        saveLoadManager.saveOutside(selectedFile, codeArea.getText());
    }

    public void load(String selectedFile) throws IOException {
        String[] origCode = saveLoadManager.load(selectedFile);
        // write loaded code to codeTextArea
        for (String codeLine : origCode) {
            codeArea.append(codeLine);
        }
        // write optimized code to executionTextArea
        refreshExecutionCode();
    }

    private void refreshExecutionCode() {
        /*executionArea.setText("");
        String[] codeStrings = programCode.getCode();
        for (String codeLine : codeStrings) {
            executionArea.append(codeLine);
        }
        ramController.visualizeStack();*/
        //TODO заполнение таблицы выполнения
    }

    private void setWidgetsToHardware(CPU cpu, HashMap widMap) {
        for (RegNames regName : RegNames.values()) {
            TextWidget regWidget = (TextWidget) widMap.get(regName);
            cpu.setRegWidget(regName, regWidget);
        }
    }

   /* public void compile() {
        saveLoadManager.saveInside(codeArea.getText());
        runProgramHandler.compile();
    }*/
}
